''' User Account and Roles Models '''

import enum
import uuid

from datetime import datetime, timedelta
from passlib.context import CryptContext
from sqlalchemy import (
    Column,
    Date,
    DateTime,
    Enum,
    ForeignKey,
    Integer,
    Text,
    Unicode
    )
from sqlalchemy.orm import relationship

from .meta import Base, JsonEncoder


#: create a crypt context for password hashes
#: configured in :mod:`ordr2.security.includeme()`
#: this is not in :mod:`ordr2.security` to avoid circular imports
passlib_context = CryptContext()


# non-database models

class Role(enum.Enum):
    ''' roles of user accounts '''

    #: new user, email address was not validated
    UNVALIDATED = 'unvalidated'

    #: new user, email address validated, not activated by admin
    NEW = 'new'

    #: standard user of the system, can place and view orders
    USER = 'user'

    #: privileged user of the system, can edit orders
    PURCHASER = 'purchaser'

    #: privileged user, can edit orders, users and consumables
    ADMIN = 'admin'

    #: a user that is no longer activated
    INACTIVE = 'inactive'

    @property
    def principal(self):
        ''' returns the principal identifier of the role '''
        return 'role:' + self.value.lower()

    def __str__(self):
        ''' string representation '''
        return self.value.capitalize()


class TokenSubject(enum.Enum):
    ''' Token Subjects for changing user accounts '''

    #: validate email address of freshly registered user
    USER_REGISTRATION = 'user_registration'

    #: validate email change of active user
    CHANGE_EMAIL = 'change_email'

    #: reset a forgotten password
    RESET_PASSWORD = 'reset_password'


# database driven models

class User(Base):
    ''' A user of the application '''

    __tablename__ = 'users'

    #: primary key
    id = Column(Integer, primary_key=True)
    #: unique user name
    username = Column(Text, nullable=False, unique=True)
    #: hashed password, see :mod:`ordr2.security`
    password_hash = Column(Text, nullable=False)
    #: role of the user, see :class:`ordr2.models.account.Role`
    role = Column(Enum(Role), nullable=False)

    first_name = Column(Text, nullable=False)
    last_name = Column(Text, nullable=False)
    email = Column(Text, nullable=False, unique=True)
    date_created = Column(Date, nullable=False, default=datetime.utcnow)

    #: tokens for new user registration, email change and forgotten passwords
    tokens = relationship(
        'Token',
        back_populates='owner',
        cascade="all, delete-orphan"
        )

    @property
    def principal(self):
        ''' returns the principal identifier for the user '''
        return 'user:{}'.format(self.id)

    @property
    def role_principals(self):
        ''' returns the principal identifiers for the user's role '''
        principals = [self.role.principal]
        if self.role is Role.PURCHASER:
            # a purchaser is also a user
            principals.append(Role.USER.principal)
        elif self.role is Role.ADMIN:
            # an admin is also a purchaser and a user
            principals.append(Role.PURCHASER.principal)
            principals.append(Role.USER.principal)
        return principals

    @property
    def is_active(self):
        ''' is true if the user has an active role '''
        return self.role in (Role.USER, Role.PURCHASER, Role.ADMIN)

    def set_password(self, password):
        ''' hashes a password using :mod:`ordr2.security.passlib_context` '''
        self.password_hash = passlib_context.hash(password)

    def check_password(self, password):
        ''' checks a password against a stored password hash

        if the password algorithm is considered deprecated, the stored hash
        will be updated using the current algorithm
        '''
        ok, new_hash = passlib_context.verify_and_update(
            password,
            self.password_hash
            )

        if not ok:
            # password does not match, return False
            return False
        elif new_hash:
            # algorithm is deprecated, update hash with new algorithm
            self.password_hash = new_hash

        # password match, return True
        return True

    def issue_token(self, request, subject, payload=None):
        ''' issues a token for mail change, password reset or user verification

        :param request:
            the current request object
        :type request:
            pyramid.request.Request
        :param subject:
            what the token is used for
        :type subject:
            ordr2.models.account.TokenSubject
        :param payload:
            extra data to store with the token, must be JSON serializable
        :rtype:
            (str) unique hash to access the token
        '''
        return Token.issue(request, self, subject, payload)

    def __str__(self):
        ''' string representation '''
        return str(self.username)


class Token(Base):
    ''' Tokens for mail change, account verification and password reset '''

    __tablename__ = 'tokens'

    #: hash identifyer of the token
    hash = Column(Unicode, primary_key=True)

    #: :class:`ordr2.models.account.TokenSubject`
    subject = Column(Enum(TokenSubject), nullable=False)

    #: token expires at this date and time
    expires = Column(DateTime, nullable=False)

    #: additional data to attach to a token
    payload = Column(JsonEncoder, nullable=True)

    #: the user_id the token belongs to
    owner_id = Column(Integer, ForeignKey('users.id'))

    #: the user the token belongs to
    owner = relationship('User', back_populates='tokens')

    @classmethod
    def issue(cls, request, owner, subject, payload=None):
        ''' issues a token for mail change, password reset or user verification

        if the expiry keys for the token is not set in the app configuration,
        the token will expire in five minutes.

        to set the expiry time in the conig use `token_expiry.` prefix followed
        by the value of the token subject and a time in minutes. For example,
        to give an active user two hours time to verify an email address change
        use `token_expiry.change_email = 120`

        :param request:
            the current request object
        :type request:
            pyramid.request.Request
        :param subject:
            what the token is used for
        :type subject:
            ordr2.models.account.TokenSubject
        :param owner:
            account the token is issued for
        :type subject:
            ordr2.models.account.User
        :param payload:
            extra data to store with the token, must be JSON serializable
        :rtype:
            ordr2.models.account.Token
        '''
        settings_key = 'token_expiry.' + subject.value
        minutes = request.registry.settings.get(settings_key, 5)
        expires = datetime.utcnow() + timedelta(minutes=int(minutes))
        return cls(
            hash=uuid.uuid4().hex,
            subject=subject,
            payload=payload,
            owner=owner,
            expires=expires)
